@@ -299,9 +299,9 @@ def setup_parser():
     parser.add_argument(
         '-s', '--deploy-delay', action='store', type=float,
         help=("Time in seconds to sleep between 'deploy' commands, "
-              "to allow machine provider to process requests. This "
-              "delay is also enforced between calls to "
-              "terminate_machine"),
+              "to allow machine provider to process requests. On "
+              "terminate machines this also signals waiting for "
+              "machine removal."),
         dest="deploy_delay", default=0)
     parser.add_argument(
         '-e', '--environment', action='store', dest='juju_env',
@@ -573,7 +573,7 @@ class Git(Vcs):
         params = ["git", "rev-parse", "HEAD"]
         return self._call(params, self.err_cur_rev)
 
-    def update(self,  rev=None):
+    def update(self, rev=None):
         params = ["git", "reset", "--merge"]
         if rev:
             params.append(rev)
@@ -626,7 +626,7 @@ class Charm(object):
 
         charm_path, store_url, build = None, None, None
         name = d.get('charm', name)
-        if name.startswith('cs'):
+        if name.startswith('cs:'):
             store_url = name
         else:
             charm_path = path_join(series_path, name)
@@ -838,7 +838,8 @@ class Deployment(object):
             expanded = []
             for c in v['consumes']:
                 expanded.append((k, c))
-            rels[v.get('weight', 0)] = expanded
+            by_weight = rels.setdefault(v.get('weight', 0), [])
+            by_weight.extend(expanded)
         for k in sorted(rels):
             for r in rels[k]:
                 if check(*r):
@@ -955,7 +956,13 @@ class Deployment(object):
 
     def save(self, path):
         with open(path, "w") as fh:
-            fh.write(yaml_dump(self.data, defaul_flow_style=False))
+            fh.write(yaml_dump(self.data))
+
+    @staticmethod
+    def to_yaml(dumper, deployment):
+        return dumper.represent_dict(deployment.data)
+
+yaml.add_representer(Deployment, Deployment.to_yaml)
 
 
 class BaseEnvironment(object):
@@ -999,15 +1006,16 @@ class BaseEnvironment(object):
 #        fh.write(yaml_dump(config))
 #        fh.flush()
 
-    def _get_units_in_error(self):
+    def _get_units_in_error(self, status=None):
         units = []
-        status = self.status()
+        if status is None:
+            status = self.status()
         for s in status.get('services', {}).keys():
             for uid, u in status['services'][s].get('units', {}).items():
-                if u['agent-state'] == 'error':
+                if 'error' in u['agent-state']:
                     units.append(uid)
                 for uid, u in u.get('subordinates', {}).items():
-                    if u['agent-state'] == 'error':
+                    if 'error' in u['agent-state']:
                         units.append(uid)
         return units
 
@@ -1046,7 +1054,12 @@ class BaseEnvironment(object):
             raise RuntimeError("Can't terminate machine 0")
         params = self._named_env(["juju", "terminate-machine"])
         params.append(mid)
-        _check_call(params, self.log, "Error terminating machine %r" % mid)
+        try:
+            _check_call(params, self.log, "Error terminating machine %r" % mid)
+        except ErrorExit, e:
+            if ("machine %s does not exist" % mid) in e.error.output:
+                return
+            raise
 
     def get_service_address(self, svc_name):
         status = self.get_cli_status()
@@ -1143,11 +1156,48 @@ class PyEnvironment(BaseEnvironment):
             self, timeout, goal_state="started", watch=False, no_exit=False):
 
         max_time = time.time() + timeout
-        status = self.status()
         while max_time > time.time():
-            for s in status.get('services', {}).values():
-                for u in s.get('units', {}).values():
-                    value = u.get('state')
+            status = self.status()
+            pending = []
+            error_units = self._get_units_in_error(status)
+            errors = []
+            for s in status.get("services", {}).values():
+                for uid, u in s.get("units", {}).items():
+                    state = u.get("agent-state") or "pending"
+                    if uid in error_units:
+                        errors.append({"name": uid,
+                                       "machine": u["machine"],
+                                       "agent-state": state})
+                    elif state != goal_state:
+                        pending.append(u)
+                    for rid in u.get("relation-errors", {}).keys():
+                        errors.append({"name": uid,
+                                       "machine": u["machine"],
+                                       "agent-state":
+                                       "relation-error: %s" % rid})
+                    for sid, sub in u.get("subordinates", {}).items():
+                        state = sub.get("agent-state") or "pending"
+                        if sid in error_units:
+                            errors.append({"name": sid,
+                                           "machine": u["machine"],
+                                           "agent-state": state})
+                        elif state != goal_state:
+                            pending.append(sid)
+            if not pending and not errors:
+                break
+            if errors:
+                if no_exit:
+                    raise UnitErrors(errors)
+                else:
+                    error_units = [
+                        "unit: %s: machine: %s agent-state: %s" % (
+                            u["name"], u["machine"], u["agent-state"]
+                            )
+                        for u in errors]
+                    self.log.error(
+                        "The following units had errors:\n  %s" % (
+                            "   \n".join(error_units)))
+                    raise ErrorExit()
 
 
 class GoEnvironment(BaseEnvironment):
@@ -1204,7 +1254,7 @@ class GoEnvironment(BaseEnvironment):
 
     def reset(self,
               terminate_machines=False,
-              terminate_delay=0,
+              terminate_wait=False,
               timeout=360, watch=False):
         """Destroy/reset the environment."""
         status = self.status()
@@ -1224,19 +1274,31 @@ class GoEnvironment(BaseEnvironment):
         # The only value to not terminating is keeping the data on the
         # machines around.
         if not terminate_machines:
-            self.log.info(" *juju-core machines are not reusable for units")
+            self.log.info(
+                " warning: juju-core machines are not reusable for units")
             return
+        self._terminate_machines(status, watch, terminate_wait)
 
+    def _terminate_machines(self, status, watch, terminate_wait):
+        """Terminate all machines, optionally wait for termination.
+        """
         # Terminate machines
         self.log.debug(" Terminating machines")
+
+        # Don't bother if there are no service unit machines
+        if len(status['machines']) == 1:
+            return
+
         for mid in status['machines'].keys():
             if mid == "0":
                 continue
             self.log.debug("  Terminating machine %s", mid)
             self.terminate_machine(mid)
-            if terminate_delay:
-                self.log.debug("  Waiting for terminate delay")
-                time.sleep(terminate_delay)
+
+        if terminate_wait:
+            self.log.info("  Waiting for machine termination")
+            callback = watch and self._delta_event_log or None
+            self.client.wait_for_no_machines(None, callback)
 
     def _check_timeout(self, etime):
         w_timeout = etime - time.time()
